﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=9.8
@EndOfDesignText@
'Effectively an extension of WavRandomAccessFile so that controls for Gain, Volume and Balance can be added to control the mixer.
Sub Class_Globals
	Private ARAF As WavRandomAccessFile
	Private mRAF As RandomAccessFile
	Private mMute As Boolean
	Private mSolo As Boolean
	Private mGain As Float  = 1.0
	Private mVolume As Float  = 1.0
	Private mPlayPos As Long
	Private mLeftBalance As Float = 1.0
	Private mRightBalance As Float = 1.0
	Private mDuration As Long
	Private mDataChunkStart As Long
	Private mDataChunkEnd As Long
	Private mFilePath, mFileName As String
	Private mTrackName As String
End Sub

'Initializes the object. You can add parameters to this method if needed.
Public Sub Initialize
	
End Sub

Public Sub SetFile(FilePath As String, FileName As String) As Boolean
	Try
		'Remove existing file from this track
		If ARAF.IsInitialized And mFileName <> "" Then
			For i = 0 To TrackMixer_G.SourceFilesList.Size - 1
				Dim SF As SourceFileType = TrackMixer_G.SourceFilesList.Get(i)
				If SF.Path = mFilePath And SF.FileName = mFileName Then
					TrackMixer_G.SourceFilesList.RemoveAt(i)
				End If
			Next
			ARAF.Close
		End If
	Catch
		Log("Error closing RandomAccessFile " & LastException)
	End Try
	
	'Do this after closing previous ARAF
	mFilePath = FilePath
	mFileName = FileName
	mTrackName = FileName.SubString2(0,Max(FileName.Length,FileName.IndexOf(".")))

	'Create a new WavRandomAccessFile and add a record to the Project SourceFileList
	Dim ARAF As WavRandomAccessFile
	If ARAF.Initialize(mFilePath,mFileName) Then
		'Check that the audioformat of the file matches the project defined format
		If Project_Static.ProjectAudioFormatMatches(getAudioFormat) = False Then
			Log("ProjectAudioFormat mismatch")
			Return False
		End If
	
		Dim SF As SourceFileType
		SF.Initialize
		SF.Path = FilePath
		SF.FileName = FileName
		SF.Track = Me
		TrackMixer_G.SourceFilesList.Add(SF)
	
		'Direct pointers / copies of properties wrapped in the WavRandomAccessFile for speed of retreival
		mRAF = ARAF.RandomAccessFile
		mDuration = ARAF.Duration
		mDataChunkStart = ARAF.DataChunkStart
		mDataChunkEnd = ARAF.DataChunkEnd
	Else
		Return False
	End If
	
	Return True
End Sub

Public Sub Close
	ARAF.Close
End Sub

'Start position of the DataChunk in the Wav file
Public Sub DataChunkStart As Long
	Return mDataChunkStart
End Sub

'Dnd position of the DataChunk in the Wav file
Public Sub DataChunkEnd As Long
	Return mDataChunkEnd
End Sub

'Wav file audio format
Public Sub getAudioFormat As JavaObject
	Return ARAF.AudioFormat
End Sub

'Direct access to the RAF Object wrapped in the AudioRandomAccess File for speed of retreival
Public Sub getRandomAccessFile As RandomAccessFile
	Return mRAF
End Sub

'Get set the gain for this track
Public Sub getGain As Float
	Return mGain
End Sub

Public Sub setGain(Gain As Float)
	mGain = Gain
End Sub

Public Sub setVolume(Volume As Float)
	mVolume = Volume
End Sub

Public Sub getVolume As Float
	Return mVolume
End Sub

'Get/Set the play position for this track
Public Sub setPlayPos(Pos As Long)
	mPlayPos = Pos
End Sub

Public Sub getPlayPos As Long
	Return mPlayPos
End Sub

'Get the Track Duration in Seconds
Public Sub getDuration As Long
	Return mDuration
End Sub

'Get/Set the Mute state for this track
Public Sub setMute(Mute As Boolean)
	mMute = Mute
End Sub

Public Sub getMute As Boolean
	Return mMute
End Sub

Public Sub setSolo(Solo As Boolean)
	mSolo = Solo
	TrackMixer_G.setsolo(Me,Solo)
End Sub

Public Sub getSolo As Boolean
	Return mSolo
End Sub

'Get/Set the Left balance for this track
Public Sub setLeftBalance(Val As Float)
	mLeftBalance = Val
End Sub

Public Sub getLeftBalance As Float
	Return mLeftBalance
End Sub

'Get/Set the Right balance for this track
Public Sub setRightBalance(Val As Float)
	mRightBalance = Val
End Sub

Public Sub getRightBalance As Float
	Return mRightBalance
End Sub

Public Sub getName As String
	Return mTrackName
End Sub

