﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=StaticCode
Version=9.8
@EndOfDesignText@
'Threaded version of TrackMixer, allows using a lower SampleRate or playing more tracks.
Sub Process_Globals
	Type SourceFileType(Path As String, FileName As String, Track As MixerTrack)
	Private mSourceFilesList As List
	Private fx As JFX
	Private mTracks As List
	Private OutDevice As JavaObject
	Private SDLW As SourceDataLineWrapper
	Private MaxBufferSize As Int = 2048
	Private BufferSize As Int
	Private StopPlaying As Boolean
	Private mPause As Boolean
	Private ReportPos As Boolean
	Private PosCallback As Object
	Private PosEventName As String
	Private ReportClip As Boolean
	Private ClipCallback As Object
	Private ClipEventName As String
	Private ReportSound As Boolean
	Private SoundCallback As Object
	Private SoundEventName As String
	Private SoundThreshold As Short = 500
	Private PlayPos As Long
	Private misPlaying As Boolean
	Private mMasterGain As Float = 1.0
	Private WarningClipValue As Short = 29000 'arbitrary value after which clipping is likely to occur adding digital artefacts to the sound
	Private FirstTime As Boolean
	
	Private TH As Thread				'Ignore

End Sub


'Initializes the object. You can add parameters to this method if needed.
Public Sub Initialize
	mTracks.Initialize
	TH.Initialise("TH")
	mSourceFilesList.Initialize
End Sub

Public Sub Version As String
	Return "Threaded"
End Sub

Public Sub SourceFilesList As List
	If mSourceFilesList.IsInitialized = False Then mSourceFilesList.Initialize
	Return mSourceFilesList
End Sub

Public Sub AddPositionListener(Callback As Object,EventName As String)
	PosCallback = Callback
	PosEventName = EventName
	If SubExists(PosCallback,PosEventName & "_PosEvent") Then ReportPos = True
End Sub

Public Sub AddClipListener(Callback As Object,EventName As String)
	ClipCallback = Callback
	ClipEventName = EventName
	If SubExists(ClipCallback,ClipEventName & "_ClipEvent") Then ReportClip = True
End Sub

Public Sub AddSoundListener(Callback As Object,EventName As String)
	SoundCallback = Callback
	SoundEventName = EventName
	If SubExists(SoundCallback,SoundEventName & "_SoundEvent") Then ReportSound = True
End Sub

Public Sub AddTrack(MT As MixerTrack)
	mTracks.Add(MT)
	
	If OutDevice.IsInitialized = False Then
		OutDevice = jAudioTrack2_Utils.GetSourceDataLine(MT.AudioFormat)
		SDLW.Initialize(OutDevice)
	End If
End Sub

Public Sub RemoveTrack(MT As MixerTrack)
	mTracks.RemoveAt(mTracks.IndexOf(MT))
	For i = 0 To mSourceFilesList.Size - 1
		Dim SF As SourceFileType = mSourceFilesList.Get(i)
		If SF.Track = MT Then
			mSourceFilesList.RemoveAt(i)
			SF.Track.close
			Exit
		End If
	Next
End Sub

Public Sub Close
	Stop
	For Each SF As SourceFileType In mSourceFilesList
		SF.Track.close
	Next
End Sub

Public Sub RemoveAllTracks
	For Each SF As SourceFileType In mSourceFilesList
		SF.Track.close
	Next
	mTracks.Clear
	mSourceFilesList.Clear
End Sub

Public Sub Stop
	StopPlaying = True
End Sub

Public Sub Pause
	mPause = True
End Sub

Public Sub Play
	If misPlaying Then Return
	StopPlaying = False
	mPause = False
	TH.Start(Me,"Play_Do",Null)
End Sub

Private Sub Play_Do 		'Ignore
	Log("Playing Threaded")
	misPlaying = True
	
	'check there are tracks to play
	If mTracks.Size = 0 Then
		misPlaying = False
		Return
	End If
	
	'Required library
	Dim BC As ByteConverter
	BC.LittleEndian = True
	
	'Working buffer size
	BufferSize = Min(MaxBufferSize,SDLW.GetBufferSize)
	
	'Start the source data line
	SDLW.Start

	'Working buffer
	Dim Dat(BufferSize) As Byte
	Dim ThisBufferSize As Int

	'Sample values
	Dim NewSample, Mixed As Short

	'Running total for each buffer operation (all tracks)
	Dim TotalRead As Int
	
	'Position indexes
	Dim i,MixBufferi As Int
	Dim Len As Long
	Dim MaxValue As Short
	
	Dim SoundMap As Map
	SoundMap.Initialize
	
	FirstTime = True
	
	Do While True
		'Maximum length of the last buffer read from each track
		Len = 0
		TotalRead = 0
		
		'Define / Clear the mix buffer for the nexr iteration
		Dim MixBuffer(BufferSize / 2) As Short
		
		
		'Mixing Process
		'For each track:
		'Read each loaded file in sequence as NewSample(2 bytes per sample Left channel then Right channel),
		'Aquire each 16bit sample and convert to float
		'Check the level of each track is not too high (WarningClipValue is an arbitrary value of 29000 & -29000)
		'Adjust the level of the left and right samples by the Track left and right balance values.
		'Add The NewSample to the corresponding value in the MixBuffer(all tracks already processed) into the variable Mixed
		'Check the level of the mixed value is not too high (WarningClipValue is an arbitrary value of 29000 & -29000)
		'Add the Mixed value to the corresponding position of the shortbuffer MixBuffer.
		
		'Once all tracks are added to the MixBuffer:
		'Convert the Short Mixbuffer to bytes and write it to the SourceDataLine
		'Check if the Stop or pause flags are set and exit if needed
		'If the total bytes read from each track is 0, then playing of all tracks is complete so exit.
		'Tidy up setting appropriate flags, stopping the SourceDataLine and return from the sub
		
		For Each Track As MixerTrack In mTracks
			MaxValue = 0
			'Avoid reading past the end of the Data Chunk in case the Data Chunk is not the last chunk in the file.
			ThisBufferSize = Max(0, Min(BufferSize , Track.DataChunkEnd - Track.PlayPos))
			
			
			TotalRead = TotalRead + Track.RandomAccessFile.ReadBytes(Dat,0,ThisBufferSize, Track.DataChunkStart + Track.PlayPos)
			Len = Max(Len,Dat.Length)
			
			'Play position for each track
			Track.PlayPos = Track.PlayPos + Dat.Length
			'Check if track is muted
			If Track.Mute = False Then
				For i = 0 To Dat.Length - 1 Step 2
					MixBufferi = i / 2
					'Convert 16 bits to short so they can be added toGether
					NewSample = BC.ShortsFromBytes(Array As Byte(Dat(i),Dat(i+1)))(0) * Track.Gain * Track.Volume * mMasterGain
					
					If NewSample > WarningClipValue Then
						If ReportClip Then TH.RunOnGuiThread("Report_Clip",Array(Track,NewSample))
					End If
					If NewSample < -WarningClipValue Then
						If ReportClip Then TH.RunOnGuiThread("Report_Clip",Array(Track,NewSample))
					End If
					
					'Deal with Balance of the stereo track
					If i Mod 4 = 0 Then
						NewSample = NewSample * Track.LeftBalance
					Else
						NewSample = NewSample * Track.RightBalance
					End If
					
					'Add the NewSample to the MixBuffer corresponding value
					Mixed = (MixBuffer(MixBufferi) + NewSample)

					'Hard limit the output
					If Mixed > WarningClipValue Then
						If ReportClip Then TH.RunOnGuiThread("Report_Clip",Array(Null,Mixed))
					End If
					If Mixed < -WarningClipValue Then
						If ReportClip Then TH.RunOnGuiThread("Report_Clip",Array(Null,Mixed))
					End If
					
					'Update the current Mixbuffer
					MixBuffer(MixBufferi) = Mixed
					MaxValue = Max(MaxValue,NewSample)
				Next
				
			End If
			If ReportSound And MaxValue > SoundThreshold Then SoundMap.Put(Track,MaxValue)
			
			'Exit if flag is set
			If StopPlaying Then Exit
		Next

		'Report current position if requested
		If ReportPos Then TH.RunOnGuiThread("Report_Position",Array(PlayPos))
		
		'Update play position for reporting (Maximum value)
		PlayPos = PlayPos + Len
		
		
		'Convert the mixed MixBuffer to bytes and write it to the SourceDataLine
		SDLW.Write(BC.ShortsToBytes(MixBuffer),0,MixBuffer.Length * 2)
		
		
		If SoundMap.Size > 0 Then 
			'Clone the soundmap so it can be cleared for the next iteration.
			Dim SoundMap1 As Map
			SoundMap1.Initialize
			SoundMap1.As(JavaObject).RunMethod("putAll",Array(SoundMap))
			SoundMap.Clear
			TH.RunOnGuiThread("Report_Sound",Array(IIf(FirstTime,True,False),SoundMap1))
		End If
		FirstTime = False
		'Exit if flag is set
		If StopPlaying Or mPause Then Exit
		
		'Will occur when all RandomAccess files are exhausted.
		If TotalRead = 0 Then Exit
	Loop
	
	If StopPlaying Or TotalRead = 0 Then
		'Reset play position to beginning. If playing is paused, this doesn't happen.
		SetPlayPos(0)
		
		'Report current position if requested.  In this app it resets the position slider play position
		If ReportPos Then TH.RunOnGuiThread("Report_Position",Array(PlayPos))
	End If
	
	'Stop and clear the Source Data Line
	SDLW.Stop
	SDLW.Drain
	SDLW.flush
	
	'Reset the playing flag
	misPlaying = False
End Sub

Private Sub TH_Ended(endedOK As Boolean, error As String) 'The thread has terminated. If endedOK is False error holds the reason for failure
	If endedOK = False Then
		Log("Thread Ended OK " & endedOK)
		Log("Error " & error)
	End If
End Sub

Private Sub Report_Position(Pos As Long)
	CallSubDelayed2(PosCallback,PosEventName & "_PosEvent",Pos)
End Sub

Private Sub Report_Clip(Track As MixerTrack,MixedValue As Short)
	CallSubDelayed3(ClipCallback,ClipEventName & "_ClipEvent",Track,MixedValue)
End Sub

Private Sub Report_Sound(tFirstTime As Boolean,SoundMap As Object)
	Dim SoundMap1 As Map = SoundMap
	If Not(tFirstTime) Then Sleep(500)
	For Each Track As MixerTrack In SoundMap1.Keys
		Dim MaxVal As Short = SoundMap1.Get(Track)
		CallSubDelayed3(SoundCallback,SoundEventName & "_SoundEvent",Track,MaxVal)
	Next
End Sub

Public Sub SetSolo(MT As MixerTrack, Solo As Boolean)
	
	For Each T As MixerTrack In mTracks
		If T = MT Then
			T.Mute = False
		Else
			If Solo Then
				T.Mute = True
			Else
				T.Mute = False
			End If
		End If
	Next

	
End Sub

Public Sub SetSoundThreshold(Threshold As Short)
	SoundThreshold = Threshold
End Sub

Public Sub GetSoundThreshold As Short
	Return SoundThreshold
End Sub

Public Sub SetMaxBufferSize(tBufferSize As Int)
	MaxBufferSize = tBufferSize
End Sub

Public Sub GetMaxBufferSize As Int
	Return MaxBufferSize
End Sub

Public Sub SetWarningClipValue(Value As Short)
	WarningClipValue = Value
End Sub

Public Sub GetWarningClipValue As Short
	Return WarningClipValue
End Sub

Public Sub IsPlaying As Boolean
	Return misPlaying
End Sub

Public Sub SetMasterGain(Gain As Float)
	mMasterGain = Gain
End Sub
'Get/Set Master gain
Public Sub GetMasterGain As Float
	Return mMasterGain
End Sub

Public Sub GetTracks As List
	Return mTracks
End Sub

'Get the sourceDataLine
Public Sub GetSourceDataLine As SourceDataLineWrapper
	Return SDLW
End Sub

'Get/Set the play position
Public Sub SetPlayPos(Pos As Long)
	PlayPos = Pos
	For Each SF As SourceFileType In mSourceFilesList
		SF.Track.PlayPos = Pos
	Next
End Sub

Public Sub GetPlayPos As Long
	Dim Pos As Long
	For Each SF As SourceFileType In mSourceFilesList
		Pos = Max(Pos,SF.Track.PlayPos)
	Next
	Return Pos
End Sub

'Get the play duration in Milliseconds
Public Sub GetMilliSecondDuration As Long
	Dim Dur As Long
	For Each SF As SourceFileType In mSourceFilesList
		Dur = Max(Dur,SF.Track.Duration)
	Next
	Return Dur * 1000
End Sub

'Get/Set the millisecond play position
Public Sub GetMillisecondPosition As Long
	Return PlayPos / (Project_Static.ProjectSampleRateInHz * Project_Static.ProjectChannelConfig * (Project_Static.ProjectSampleSizeInBits / 8)) * 1000
End Sub

Public Sub SetMillisecondPosition(mSeconds As Long)
	Dim NewPos As Long = Floor((mSeconds / 1000) * Project_Static.ProjectSampleRateInHz * Project_Static.ProjectChannelConfig) * (Project_Static.ProjectSampleSizeInBits / 8)
	PlayPos = NewPos
	
	'Set on each track
	For Each SF As SourceFileType In mSourceFilesList
		SF.Track.PlayPos = PlayPos
	Next
End Sub